from enum import Enum
import json
import re
import textwrap
from typing import Any, Dict, Optional, Sequence, cast

from jinja2 import Environment, PackageLoader, Template
from WDL import (
    Call,
    Conditional,
    Decl,
    Document,
    Env,
    Expr,
    Scatter,
    StructTypeDef,
    Task,
    Workflow,
    WorkflowNode,
    WorkflowSection,
)
from WDL.Tree import DocImport

from wdlkit.ast import WorkflowBodyGraph


class WdlTemplate(Enum):
    """
    Enumeration of Jinja2 templates that are used to render a WDL document.
    """

    CALL = "call"
    COMMAND = "command"
    CONDITIONAL = "conditional"
    DECLARATIONS = "declarations"
    DOCUMENT = "document"
    IMPORTS = "imports"
    INPUTS = "inputs"
    META = "meta"
    OUTPUTS = "outputs"
    PARAMETER_META = "parameter_meta"
    RUNTIME = "runtime"
    SCATTER = "scatter"
    STRUCT = "struct"
    TASK = "task"
    WORKFLOW = "workflow"
    WORKFLOW_BODY = "workflow_body"


def _create_jinja2_env():
    env = Environment(loader=PackageLoader("wdlkit"))
    env.filters["dedent"] = textwrap.dedent
    return env


class Formatter:
    """
    WDL document formatter.
    """

    _JINJA2_ENV = _create_jinja2_env()
    _JSON_ENCODER = json.JSONEncoder(indent=2)
    _UNQUOTE_RE = re.compile(r"^(\s+)\"(.+?)\":", re.M)
    _UNQUOTE_REPL = r"\1\2:"

    @classmethod
    def _get_template(cls, tmpl: WdlTemplate) -> Template:
        return Formatter._JINJA2_ENV.get_template(f"{tmpl.value}.wdl")

    @classmethod
    def _render_template(cls, tmpl: WdlTemplate, **kwargs) -> str:
        return cls._get_template(tmpl).render(**kwargs)

    @classmethod
    def _format_jsonlike(cls, value: Dict[str, Any]) -> Dict[str, str]:
        # format as JSON and unquote hash keys
        return dict(
            (
                key,
                Formatter._UNQUOTE_RE.subn(
                    Formatter._UNQUOTE_REPL, Formatter._JSON_ENCODER.encode(val)
                )[0],
            )
            for key, val in value.items()
        )

    @classmethod
    def format_document(
        cls, uri: str, doc: Document, contents: Optional[Dict[str, str]] = None
    ) -> Dict[str, str]:
        """
        Formats a WDL document and all its imports.

        Args:
            uri: URI or local path of the main WDL document.
            doc: The WDL AST.
            contents: Used to pass the mapping of URI to formatted contents to
                recursive calls on imports. Callers should not need to use this.

        Returns:
            Mapping of URI to formatted document.
        """
        if contents is None:
            contents = {}

        contents[uri] = cls._render_template(
            WdlTemplate.DOCUMENT,
            imports=cls.format_imports(doc.imports),
            structs=[cls.format_struct(s) for s in doc.struct_typedefs or ()],
            workflow=cls.format_workflow(doc.workflow) if doc.workflow else None,
            tasks=[cls.format_task(task) for task in (doc.tasks or ())],
        )

        if doc.imports:
            for imp in doc.imports:
                cls.format_document(imp.uri, imp.doc, contents)

        return contents

    @classmethod
    def format_workflow(cls, value: Workflow) -> str:
        return cls._render_template(
            WdlTemplate.WORKFLOW,
            name=value.name,
            inputs=cls.format_input(value.inputs or ()),
            body=cls.format_workflow_body(value.body),
            outputs=cls.format_output(value.outputs or ()),
            meta=cls.format_meta(value.meta),
            parameter_meta=cls.format_parameter_meta(value.parameter_meta),
        )

    @classmethod
    def format_workflow_body(cls, value: Sequence[WorkflowNode]) -> str:
        graph = WorkflowBodyGraph(*value)
        decls = []
        body = []

        for node in graph.dependency_order():
            if isinstance(node, Decl):
                decls.append(node)
                continue

            if decls:
                body.append(cls.format_declarations(decls))
                decls = []

            if isinstance(node, Call):
                body.append(cls.format_call(node))
            elif isinstance(node, WorkflowSection):
                body.append(cls.format_section(cast(WorkflowSection, node)))

        if decls:
            body.append(cls.format_declarations(decls))

        return cls._render_template(WdlTemplate.WORKFLOW_BODY, body=body)

    @classmethod
    def format_task(cls, value: Task) -> str:
        return cls._render_template(
            WdlTemplate.TASK,
            name=value.name,
            inputs=cls.format_input(value.inputs or ()),
            post_inputs=cls.format_declarations(value.postinputs),
            command=cls.format_command(value.command),
            outputs=cls.format_output(value.outputs),
            runtime=cls.format_runtime(value.runtime),
            meta=cls.format_meta(value.meta),
            parameter_meta=cls.format_parameter_meta(value.parameter_meta),
        )

    @classmethod
    def format_imports(cls, imports: DocImport) -> str:
        return cls._render_template(WdlTemplate.IMPORTS, imports=imports)

    @classmethod
    def format_struct(cls, struct: Env.Binding[StructTypeDef]) -> str:
        return cls._render_template(WdlTemplate.STRUCT, struct=struct)

    @classmethod
    def format_input(cls, inputs: Sequence[Decl]) -> str:
        return cls._render_template(WdlTemplate.INPUTS, inputs=inputs)

    @classmethod
    def format_output(cls, outputs: Sequence[Decl]) -> str:
        return cls._render_template(WdlTemplate.OUTPUTS, outputs=outputs)

    @classmethod
    def format_declarations(cls, decls: Sequence[Decl]) -> str:
        return cls._render_template(WdlTemplate.DECLARATIONS, decls=decls)

    @classmethod
    def format_meta(cls, meta: Dict[str, Any]) -> str:
        return cls._render_template(WdlTemplate.META, meta=cls._format_jsonlike(meta))

    @classmethod
    def format_parameter_meta(cls, parameter_meta: Dict[str, Any]) -> str:
        return cls._render_template(
            WdlTemplate.PARAMETER_META,
            parameter_meta=cls._format_jsonlike(parameter_meta)
        )

    @classmethod
    def format_call(cls, call: Call) -> str:
        return cls._render_template(WdlTemplate.CALL, call=call)

    @classmethod
    def format_section(cls, section: WorkflowSection) -> str:
        subgraph_content = cls.format_workflow_body(section.body)

        if isinstance(section, Conditional):
            return cls.format_conditional(section, body=subgraph_content)
        elif isinstance(section, Scatter):
            return cls.format_scatter(section, body=subgraph_content)

    @classmethod
    def format_conditional(cls, conditional: Conditional, body: str) -> str:
        return cls._render_template(
            WdlTemplate.CONDITIONAL, conditional=conditional, body=body
        )

    @classmethod
    def format_scatter(cls, scatter: Scatter, body: str) -> str:
        return cls._render_template(WdlTemplate.SCATTER, conditional=scatter, body=body)

    @classmethod
    def format_command(cls, command: Expr.String) -> str:
        return cls._render_template(WdlTemplate.COMMAND, command=command)

    @classmethod
    def format_runtime(cls, runtime: Dict[str, Expr.Base]) -> str:
        return cls._render_template(WdlTemplate.RUNTIME, runtime=runtime)



from pathlib import Path
from typing import Optional
from urllib3.util import parse_url


def get_local_path(
    uri: str, parent: Optional[Path] = None, overwrite: bool = True
) -> Path:
    parsed = parse_url(uri)

    if parsed.scheme not in {None, "", "file"}:
        return (parent or Path.cwd()) / Path(parsed.path).name
    else:
        path = Path(uri)

        if parent:
            return parent / path.name
        elif overwrite:
            return path.absolute()
        else:
            raise ValueError(f"Cannot overwrite local path {uri}")



from pathlib import Path
from typing import Iterator, Sequence

import networkx as nx
from networkx.algorithms.dag import lexicographical_topological_sort
from WDL import Call, Conditional, Decl, Scatter, WorkflowNode, load
from WDL.CLI import read_source


HEAD_ID = "__HEAD__"
NODE_TYPE_ORDERING = {
    Decl: 1,
    Call: 2,
    Conditional: 3,
    Scatter: 4,
}


class WorkflowBodyGraph(nx.DiGraph):
    def __init__(self, *nodes: WorkflowNode):
        super().__init__()
        self.add_node(HEAD_ID, key=(0, 0))
        self._index = 1

        if nodes:
            self.add_workflow_nodes(*nodes)

    def add_workflow_nodes(self, *nodes: WorkflowNode):
        """
        Adds `WorkflowNode`s to the graph.

        Args:
            nodes: `WorkflowNode`s
        """
        for node in nodes:
            for dep in node.workflow_node_dependencies or [HEAD_ID]:
                if not self.has_node(dep):
                    self.add_node(dep, index=self._index)
                    self._index += 1

                self.add_edge(dep, node.workflow_node_id)

            node_attrs = {
                "workflow_node": node,
                "key": (NODE_TYPE_ORDERING[type(node)], self._index),
            }

            if not self.has_node(node.workflow_node_id):
                self.add_node(node.workflow_node_id, index=self._index, **node_attrs)
                self._index += 1
            else:
                attrs = self.nodes[node.workflow_node_id]

                if "workflow_node" in attrs:
                    raise ValueError("Graph has cycle or duplicate node")

                self.nodes[node.workflow_node_id].update(node_attrs)

                if "index" not in attrs:
                    attrs["index"] = self._index
                    self._index += 1

    def dependency_order(self) -> Iterator[WorkflowNode]:
        """
        Sorts nodes in topological order, with ties broken by 1. the type of node, and
        2. the order in which the node was added.

        Returns:
            An `Iterator` over the `WorkflowNode`s in the graph.
        """
        prune = [key for (key, value) in self.nodes.items() if "key" not in value]

        if prune:
            graph = self.copy()

            for nid in prune:
                graph.remove_node(nid)
        else:
            graph = self

        itr = iter(
            lexicographical_topological_sort(
                graph, key=lambda nid: self.nodes[nid]["key"]
            )
        )
        first = next(itr)

        if first != HEAD_ID:
            raise ValueError(f"Expected first node to be HEAD, instead got {first}")

        for node_id in itr:
            yield self.nodes[node_id]["workflow_node"]


def load_document(uri: str, import_dir: Sequence[Path] = ()):
    return load(uri, path=import_dir, read_source=read_source, check_quant=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wdlkit']

package_data = \
{'': ['*'], 'wdlkit': ['templates/*']}

install_requires = \
['autoclick>=0.8.1,<0.9.0', 'miniwdl>=0.7.0,<0.8.0', 'networkx>=2.4,<3.0']

entry_points = \
{'console_scripts': ['wdlkit = wdlkit.__main__:wdlkit']}

setup_kwargs = {
    'name': 'wdlkit',
    'version': '0.0.post14',
    'description': 'WDL development tools',
    'long_description': None,
    'author': 'John Didion',
    'author_email': 'jdidion@dnanexus.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Optional, Sequence

import autoclick

from wdlkit.ast import load_document
from wdlkit.formatter import Formatter
from wdlkit.utils import get_local_path
from wdlkit.version import version


@autoclick.group(version=version)
def wdlkit():
    pass


@wdlkit.command("format")
def format_wdl(
    uri: str,
    import_dir: Sequence[Path] = (),
    wdl_version: str = "1.0",
    output_dir: Optional[Path] = None,
    overwrite: bool = False,
):
    """
    Format a WDL file and its imports.

    Args:
        uri: Path or URI (file:// or http(s)://) to the main WDL file.
        import_dir: Directory in which to look for imports. May be specified multiple
            times.
        wdl_version: WDL version to generate. Currently, only 1.0 is supported.
        output_dir: Directory in which to output formatted WDL files. If not specified,
            the input files are overwritten.
        overwrite: Whether to overwrite the input WDL file(s) with the reformatted
            versions. This option is ignored if `output_dir` is specified. Currently,
            this option is disabled by default because WDLkit is in development and
            there is a risk that this command might corrupt the WDL files. Once
            WDLkit reaches v1.0, this option will change to be enabled by default.
    """
    if wdl_version != "1.0":
        raise ValueError("Only WDL v1.0 is supported currently")

    import_dir_set = set(import_dir)
    import_dir_set.add(get_local_path(uri).parent)

    doc = load_document(uri, import_dir)

    formatted = Formatter.format_document(uri, doc)

    for uri, contents in formatted.items():
        with open(get_local_path(uri, output_dir, overwrite), "wt") as out:
            out.write(contents)


if __name__ == "__main__":
    wdlkit()
